class {{ name }}({% if bases is string %}{{ bases }}{% else %}{{ bases | join(', ') }}{% endif %}):
    {% if description %}
    """
    {{ description | indent(width=4) }}
    """
    {% endif -%}
    {% if meta %}
    linkml_meta: ClassVar[LinkMLMeta] = LinkMLMeta({{ meta | pprint | indent(width=8) }})

    {% endif %}
    {%  if attributes or validators %}
        {% if attributes %}
            {% for attr in attributes.values() %}
    {{ attr }}
            {% endfor -%}
        {% endif %}
        {% if validators %}
            {% for validator in validators.values() %}

    {{ validator }}
            {% endfor -%}
        {% endif %}
    {% else %}
    pass
    {% endif %}
