# Enum: {{ gen.name(element) }} {% if element.deprecated %} <span style="color: red;"><strong> (DEPRECATED) </strong></span> {% endif %}

{% if element.description %}
{% set element_description_lines = element.description.split('\n') %}
{% for element_description_line in element_description_lines %}
_{{ element_description_line }}_
{% endfor %}
{% endif %}

URI: {{ gen.uri_link(element) }}

{% if element.permissible_values -%}
## Permissible Values

| Value | Meaning | Description |
| --- | --- | --- |
{% for pv in element.permissible_values.values() -%}
| {{ pv.text }} | {{ pv.meaning }} | {{ pv.description|enshorten }} |
{% endfor %}
{% else %}
_This is a dynamic enum_
{% endif %}

{% set slots_for_enum = schemaview.get_slots_by_enum(element.name) %}
{% if slots_for_enum is defined and slots_for_enum|length > 0 -%}
## Slots

| Name | Description |
| ---  | --- |
{% for s in schemaview.get_slots_by_enum(element.name) -%}
| {{ gen.link(s) }} | {{ s.description|enshorten }} |
{% endfor %}
{% endif %}

{% include "common_metadata.md.jinja2" %}

## LinkML Source

<details>
```yaml
{{ gen.yaml(element) }}
```
</details>
