# Type: {{ gen.name(element) }} {% if element.deprecated %} <span style="color: red;"><strong> (DEPRECATED) </strong></span> {% endif %}

{% if element.description %}
{% set element_description_lines = element.description.split('\n') %}
{% for element_description_line in element_description_lines %}
_{{ element_description_line }}_
{% endfor %}
{% endif %}

URI: {{ gen.uri_link(element) }}

{{ gen.bullet(element, "base") }}
{{ gen.bullet(element, "uri") }}
{{ gen.bullet(element, "repr") }}
{{ gen.bullet(element, "typeof") }}
{{ gen.bullet(element, "pattern", backquote=True) }}
{% if gen.number_value_range(element) %}
* Numeric Value Range: {{ gen.number_value_range(element) }}
{% endif %}

{% include "common_metadata.md.jinja2" %}
