{#-
  Jinja2 Template for imports used for a Pandera model Python module
-#}
import pandera.polars as pla
from pandera.api.polars.types import PolarsData
import polars as pl
from typing import Optional, List, Dict

from pandera.typing import (
    Index,
    DataFrame,
    Series
)
from pandera.engines.polars_engine import (
    DateTime,
    Date,
    Time,
    Enum,
    Struct,
    Object
)
