{% set lifetime_needed = needs_lifetime %}
    fn {{ name }}{% if lifetime_needed %}<'a>{% endif %}(&{% if lifetime_needed %}'a {% endif %}self) -> {{ type_getter }};
    // fn {{ name }}_mut(&mut self) -> &mut {{ type_getter }};
{% if type_bound %}
    // fn set_{{ name }}<E>(&mut self, value: {{ type_setter }}) where E: {{ type_bound }};
{% else %}
    // fn set_{{ name }}(&mut self, value: {{ type_setter }});
{% endif %}
